from busy.command.command import QueueCommand


class DropCommand(QueueCommand):
    """Move an item to the end of the todo collection of a queue"""

    name = 'drop'
    key = 'r'
    default_criteria = [1]

    @QueueCommand.wrap
    def execute(self):
        collection = self.storage.get_collection(self.queue)
        lolist, hilist = self.collection.split(self.selection)
        self.collection.data = hilist + lolist
        if len(lolist) == 1:
            self.status = f"Dropped: {str(lolist[0])}"
        elif lolist:
            self.status = f"Dropped {str(len(lolist))} Items"
        else:
            self.status = "Dropped nothing"


class PopCommand(QueueCommand):
    """Move an item to the beginning of the collection"""

    name = 'pop'
    key = 'o'
    default_criteria = ['-']

    @QueueCommand.wrap
    def execute(self):
        collection = self.storage.get_collection(self.queue)
        hilist, lolist = self.collection.split(self.selection)
        self.collection.data = hilist + lolist
        if len(hilist) == 1:
            self.status = f"Popped: {str(hilist[0])}"
        elif hilist:
            self.status = f"Popped {str(len(hilist))} Items"
        else:
            self.status = "Popped nothing"



# A class family is a set of class definitions that use single inheritance
# (each subclass inherits from only one parent) and often multiple inheritance
# (subclasses can inherit from subclasses). So it's a hierarchy of classes,
# with one super-parent at the top.
#
# We offer a way for members of the family to declare themselves simply by
# living in the right package location. Then those classes can be instantiated
# using keys or names, without having to be specifically called. The members
# act independently of each other.
#
# What we get, after importing everything and loading it all, is essentially a
# little database of classes, where class-level properties become keys for
# looking up member classes. So, for example, we can have a family of commands,
# and use a command string to look up the right command.
#
# Ultimately, the super-parent of the family -- the class at the top of the
# hierarchy -- holds the database, actually a list, in the property called
# "family". So that class can be queried to find appropriate family member
# classes or instances thereof.
#
# This utility provides functions for importing family members, loading the
# "families" property of the super-parent, and querying the family.
#
# In the process of loading and querying the class family, we need to *avoid*
# inheritance of attributes. There might be abstract intermediary classes that
# don't want to play. So we use __dict__ to ensure we're only seeing the
# atttributes that are defined on that specific class.


from importlib import import_module
from inspect import getmodule
from pathlib import Path
from re import match


class ClassFamily:

    # Return an attribute of this specific class, not inherited.

    @classmethod
    def get_member_attr(self, attribute):
        if attribute in self.__dict__:
            return self.__dict__[attribute]

    # True or False: This class has all the attributes provided. Return True if
    # no attributes provided. Use __subclasses__ so we only look at direct
    # descendents, and __dict__ so we only look at attributes defined here (not
    # inherited).

    @classmethod
    def has_member_attrs(self, *attributes):

        # Do I have all the attributes?
        matches = [a in self.__dict__ for a in attributes]

        # Python's all() function returns True if the list is empty. So we kill
        # two birds with one stone here - get a hit if there are no attributes,
        # and avoid the empty-matches problem.
        return all(matches + list(attributes))

    # Return a set of myself and all my descendents that have certain
    # attributes. If no attribute names are provided, return the entire family.

    @classmethod
    def family_members(self, *attributes):

        # Put myself into a set if I qualify
        hits = {self} if self.has_member_attrs(*attributes) else set()

        # Go through my subclasses
        for kid in self.family_children():

            # Call the same function on each subclass
            hits |= kid.family_members(*attributes)

        # Send it back
        return hits

    @classmethod
    def family_attrs(self, attribute):
        """
        Return a set of all the values of a specific attribute that exist in
        the family. The set avoids repetition of values in the result.
        """
        # Put myself into a set if I qualify
        if self.has_member_attrs(attribute):
            values = {self.get_member_attr(attribute)}
        else:
            values = set()

        # Go through my subclasses
        for kid in self.family_children():

            # Call the same function on each subclass
            values |= kid.family_attrs(attribute)

        # Send it back
        return values

    # Return a specific family member that has a specified value of an
    # attribute. Assume that there is only one, so return it as soon as it's
    # found.

    @classmethod
    def family_member(self, attribute, value):

        # See if I qualify, and if so, return myself. Use __dict__ rather than
        # hasattr() and getattr() to refer only to myself, rather than
        # inherited attributes.
        if attribute in self.__dict__:
            if self.__dict__[attribute] == value:
                return self

        # Go through the subclasses, doing the same. The return statement will
        # exit the loop, so we get out as soon as we find one, without having
        # to evaluate others.
        for kid in self.family_children():
            found = kid.family_member(attribute, value)
            if found:
                return found

    # Import the whole family, so they show up as subclasses. Then return this
    # classes subclasses. This obviates the need to import everything at
    # initialization. It happens when needed, and only once.  Family members
    # must be in modules (files) in the same directory as the parent (me).
    # Furthermore, all the files in that directory will be imported. The only
    # exception is a file with the same name as myself, to avoid circular
    # imports. This is a private method, intended to be called when needed.

    @classmethod
    def family_children(self):

        # We only need to import the family once
        if not hasattr(self, '_family_imported'):

            # Find the location of the directory and the name of the module
            module = getmodule(self)
            directory = Path(module.__file__).parent
            module_name = module.__package__

            # Go through the directory and import everything into the module
            for import_file in directory.iterdir():
                if match(r'^[^_].*\.py$', import_file.name):
                    import_module(f'{module_name}.{import_file.stem}')

            # Prevent it from needing to run again
            self._family_imported = True

        # Return the subclasses (children)
        return self.__subclasses__()



from busy.command.command import CollectionCommand


class ListCommand(CollectionCommand):
    """Show the descriptions with selection numbers, default to all"""

    name = 'list'
    # key = "l"
    default_criteria = ['1-']
    FORMATS = {
        'description': "{!s}",
        'plan_date': "{:%Y-%m-%d}",
        'done_date': "{:%Y-%m-%d}"
    }

    @CollectionCommand.wrap
    def execute(self):
        def format(item, index):
            result = f"{(index+1):>6}"
            for colname in self.collection.schema:
                format = self.FORMATS[colname]
                value = getattr(item, colname)
                result += f"  {format.format(value)}"
            return result
        return self.output_items(format, with_index=True)



from dataclasses import dataclass

from busy.command.command import QueueCommand
from busy.util import date_util


@dataclass
class DeferCommand(QueueCommand):

    timing: str = ""
    yes: bool = None
    name = 'defer'
    key = 'f'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--timing', '-t', default='tomorrow')
        self._add_confirmation_arg(parser)

    def clean_args(self):
        def update_deferral():
            self.timing = self.ui.get_string("Timing", "tomorrow")
        super().clean_args()
        if self.selection:
            items = self.collection.items(self.selection)
            self.ui.output('\n'.join([str(i) for i in items]))
            if not self.provided('timing'):
                self.timing = 'tomorrow'
            while not self.provided('yes'):
                intro = f"Defer {self.count()} to {self.date}"
                self.confirm(intro, update_deferral)

    @property
    def date(self):
        """Absolute date for deferral"""
        # Don't cache this!
        return date_util.relative_date(self.timing)

    @QueueCommand.wrap
    def execute(self):
        if not self.selection:
            self.status = "Deferred nothing"
        elif self.yes is False:
            self.status = "Defer operation unconfirmed"
        else:
            plans = self.storage.get_collection(self.queue, 'plan')
            deferred = self.collection.delete(self.selection)
            for item in deferred:
                item.plan_date = self.date
                item.state = 'plan'
            plans.extend(deferred)
            self.status = f"Deferred {self.count(deferred)} to {self.date}"



# Provide a decorator to wrap a method so that it's called within the inherited
# version of that method.
#
# Example of use:
#
# class Parent(SuperWrapper):
#     def execute(self, method, *args, **kwargs):
#         print(f"Parent execute before")
#         method(self, *args, **kwargs)
#         print(f"Parent execute after")
#
# class InBetween(Parent):
#     @Parent.wrap
#     def execute(self, method, *args, **kwargs):
#         print(f"IB execute before")
#         method(self, *args, **kwargs)
#         print(f"IB execute after")
#
# class NewChild(InBetween):
#     @InBetween.wrap
#     def execute(self, name):
#         print(f"Hello {name}")
#
# c = NewChild()
# c.execute("Jane")


class SuperWrapper:
    @classmethod
    def wrap(parent_class, method):
        def wrapper(self, *args, **kwargs):
            parent_method = getattr(parent_class, method.__name__)
            return parent_method(self, method, *args, **kwargs)
        return wrapper



from dataclasses import dataclass

from busy.command.command import QueueCommand
from busy.model.item import Item
from busy.util import date_util
from busy.util.date_util import relative_date


@dataclass
class FinishCommand(QueueCommand):
    """Mark a task as finished by moving it to the done queue"""

    yes: bool = None
    name = 'finish'
    key = 'n'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_confirmation_arg(parser)

    def clean_args(self):
        super().clean_args()
        if self.selection:
            if not self.provided('yes'):
                items = self.collection.items(self.selection)
                self.ui.output('\n'.join([str(i) for i in items]))
                intro = f"Finish {self.count()}"
                self.confirm(intro)

    @QueueCommand.wrap
    def execute(self):
        if not self.selection:
            self.status = "Finished nothing"
        elif self.yes is False:
            self.status = "Finish operation unconfirmed"
        else:
            date = date_util.today()
            dones = self.storage.get_collection(self.queue, 'done')
            plans = self.storage.get_collection(self.queue, 'plan')
            items = self.collection.delete(self.selection)
            finished = [Item(i.current, state='done',
                             done_date=date_util.today()) for i in items]
            nexts = [Item(i.next, state='todo') for i in items
                     if (i.next and not i.repeat)]
            repeats = [Item(i.description, state='plan',
                            plan_date=i.repeat_date)
                       for i in items if i.repeat]
            dones += finished
            self.collection[0:0] = nexts
            plans += repeats
            self.status = f"Finished {self.count(finished)}"
            self.status += f" / Added {self.count(nexts)}"
            self.status += f" / Planned {self.count(repeats)}"



from busy.command.command import QueueCommand
from busy.model.collection import Collection


class TagsCommand(QueueCommand):

    name = 'tags'

    @QueueCommand.wrap
    def execute(self):
        tags = set()
        for state in Collection.family_attrs('state'):
            collection = self.storage.get_collection(self.queue, state)
            for item in collection:
                tags |= item.tags
        return '\n'.join(sorted(tags))



# The Handler is the core control class for a single execution of the busy
# program. It's called when busy is run, and there are some functions to get
# things configured and started. Then one Handler is instantiated. This is the
# only place that imports the UI, Command, and Root classes. So the handler
# instance is passed around, and other classes can access what they need from
# there. Keeping the imports in one place allows the class families to be more
# modular.
#
# A Handler instance provides the following properties. For the most part, these
# properties are how commands access other class families. Remember that the
# Namespace from the Handler is the original Namespace, so use the one passed in
# to the execute() method of the Command for actual usage in an interactive UI.


import sys
from argparse import ArgumentParser
from pathlib import Path
from sys import argv

from busy.command.command import Command
from busy.storage.file_storage import FileStorage
from busy.ui.ui import UI, UserCancelError
from busy.util.python_version import confirm_python_version

PYTHON_VERSION = (3, 6, 5)


def main(*args):
    """Start everything from the command line

    args - optional entry point for testing
    """
    args = args if args else argv[1:]
    confirm_python_version(PYTHON_VERSION)
    parser = get_parser()
    namespace = parser.parse_args(args)
    try:
        handler = Handler(namespace)
        handler.execute()
    except (RuntimeError, ValueError) as error:
        print(f"Error: {str(error)}")


def get_parser():
    """Return the CLI parser"""
    parser = ArgumentParser(prog='busy')
    parser.add_argument('--root', action='store',
                        help='file system path to busy data',
                        default=None)
    subparsers = parser.add_subparsers(dest='command')
    for command in Command.family_members('name'):
        key = command.get_member_attr('key')
        aliases = [key] if key else []
        subparser = subparsers.add_parser(command.name, aliases=aliases)
        command.set_parser(subparser)
    return parser


class Handler():

    def __init__(self, namespace):
        """
        namespace - ArgParse Namespace from initial CLI entry
        """
        nsvars = vars(namespace)
        self.storage = FileStorage(nsvars.pop('root'))
        command = nsvars.pop('command')
        command = command if command else 'curses'
        command_class = Command.family_member('name', command)
        if not command_class:
            raise RuntimeError(f"Unknown command {command}")
        self.ui = UI.family_member('name', command_class.uitype)(self)
        self.command = command_class(
            storage=self.storage, ui=self.ui, **nsvars)

    def execute(self):
        """Handle the command and output its final status"""
        result = self.command.execute()
        if result:
            print(result)
        if self.command.status:
            print(self.command.status)

    # Some useful methods so other components don't need to import from
    # different class families

    def get_commands(self, *attributes):
        return Command.family_members(*attributes)

    # A method for commands within interactive UIs. Use our own storage but the UI
    # is passed in. Returns None if attribute value doesn't resolve.

    def get_command(self, attribute, value, **kwargs):
        member = Command.family_member(attribute, value)
        if member:
            args = {'storage': self.storage} | kwargs
            return member(**kwargs)



from argparse import ArgumentParser
from io import StringIO

from wizlib.rlinput import rlinput

from busy.ui.ui import TerminalUI
from busy.ui.ui import UserCancelError
from busy.ui.ui import Prompt


class ShellUI(TerminalUI):

    """The UI to execute one command passed in through the shell. There will be
    limited interactivity, if the user omits an argument on the command line,
    but otherwise this is a run and done situation.
    """

    name = "shell"

    def output(self, intro=""):
        """Print some regular output"""
        if intro:
            print(intro)

    def get_string(self, intro, default=""):
        prompt = Prompt(intro=intro)
        return rlinput(prompt=self._prompt_string(prompt), default=default)

    def get_option(self, chooser):
        """Get a choice from the user"""
        key = rlinput(prompt=self._prompt_string(chooser))
        return chooser.choice_by_key(key)

    def _prompt_string(self, prompt):
        result = []
        if prompt.intro:
            result.append(prompt.intro)
        if prompt.default:
            result.append(f"[{prompt.default}]")
        for choice in prompt.choices:
            if choice.word == prompt.default:
                continue
            pre, it, post = choice.word.partition(choice.key)
            result.append(f"{pre}({it}){post}")
        return " ".join(result) + ": "

    def write_prompt(self, prompt):
        """Output the prompt"""
        print(self._prompt_string(prompt), end="")



from ..command.command import Command


class QueuesCommand(Command):

    name = 'queues'

    @Command.wrap
    def execute(self):
        """Get the names of the queues. Cache nothing."""
        names = sorted(self.storage.queue_names)
        return '\n'.join(names)



# Various utilities related to date handling

import functools
import re
from datetime import date as Date
from datetime import datetime as DateTime
from datetime import timedelta as TimeDelta

# Isolate the today function so it can easily be patched for testing.

today = Date.today


# The string format we always use to express a date is YYYY-MM-DD

FORMAT = re.compile(r'^(\d{4})(?:\-|\/)(\d{1,2})(?:\-|\/)(\d{1,2})$')


def absolute_date(info):
    """Cast other types to datetime.date or None"""
    if not info:
        return None
    if isinstance(info, Date):
        return info
    elif isinstance(info, tuple):
        return Date(*info)
    elif isinstance(info, str) and FORMAT.match(info):
        return Date(*(int(x) for x in FORMAT.match(info).groups()))
    raise RuntimeError("Invalid date %s" % info)


# Refactored to only accept a string

def relative_date(value: str):
    """Given a string like "tomorrow" or "in 4 days", return a date"""
    value = value.lower()
    result = next(filter(None, (h(value) for h in HANDLERS)), None)
    if result:
        return result
    else:
        return absolute_date(value)


# Used during import to bring in different handlers for relative dates. Called
# during import.

HANDLERS = []


def _register(expression=r'(.*)'):
    compiled = re.compile(expression)

    def chain(func):
        @functools.wraps(func)
        def wrapper(string):
            match = compiled.match(string)
            if match:
                return func(today(), *match.groups())
        HANDLERS.append(wrapper)
        return wrapper
    return chain


# The actual different types of expressions for relative dates.

@_register(r'today')
def _match_today(today):
    return today


@_register(r'tomorrow')
def _match_tomorrow(today):
    return today + TimeDelta(1)


@_register(r'^(\d+)\s+days?$')
def _match_days(today, days):
    return today + TimeDelta(int(days))


@_register(r'^(\d+)d$')
def _match_d(today, days):
    return today + TimeDelta(int(days))


@_register(r'^(\d{1,2})$')
def _match_day_num(today, day):
    if today.day < int(day):
        return Date(today.year, today.month, int(day))
    else:
        year = today.year + (today.month // 12)
        month = (today.month % 12) + 1
        return Date(year, month, int(day))


@_register(r'(\d{1,2})(?:\-|\/)(\d{1,2})$')
def _match_month_day_num(today, month, day):
    if today < Date(today.year, int(month), int(day)):
        return Date(today.year, int(month), int(day))
    else:
        return Date(today.year + 1, int(month), int(day))


@_register(r'^(mon?|tue?s?|wed?n?e?s?|thu?r?s?|fri?|sat?u?r?|sun?)(?:day)?$')
def _match_weekday(today, weekday):
    index = ['mo', 'tu', 'we', 'th', 'fr', 'sa', 'su'].index(weekday[0:2])
    days = (6 - today.weekday() + index) % 7 + 1
    return today + TimeDelta(days)



from collections import UserList
from collections.abc import Iterable
from csv import DictReader, DictWriter
from io import IOBase

from busy.model.item import Item
from busy.util.class_family import ClassFamily
from busy.util.selector import Selector


class CollectionError(Exception):
    pass


class Collection(UserList, ClassFamily):

    # Attributes in order for reads and writes
    schema = ['description']

    def validate_item(self, item):
        if item.state != self.state:
            raise CollectionError(
                f"Incorrect state {item.state} for {self.state} collection")
        if '|' in item.description:
            raise CollectionError(
                f"Pipe symbol not permitted in item description '{item.description}'")

    def validate(self, value):
        for item in value if isinstance(value, Iterable) else [value]:
            self.validate_item(item)

    # Override list methods to make sure we've got Items of the right state.
    # There might be some methods missing here. Also set changed.

    def __setattr__(self, name, value):
        if name == 'data':
            self.validate(value)
        super().__setattr__(name, value)
        if name == 'data':
            self.changed = True

    def __setitem__(self, key, value):
        self.validate(value)
        self.changed = True
        return super().__setitem__(key, value)

    def __iadd__(self, value):
        self.validate(value)
        self.changed = True
        return super().__iadd__(value)

    def __delitem__(self, *args):
        self.changed = True
        return super().__delitem__(*args)

    def append(self, value):
        self.validate(value)
        self.changed = True
        return super().append(value)

    def extend(self, value):
        self.validate(value)
        self.changed = True
        return super().extend(value)

    def insert(self, index, value):
        self.validate(value)
        self.changed = True
        return super().insert(index, value)

    def __add__(self, value):
        self.validate(value)
        self.changed = True
        return super().__add__(value)

    def pop(self, *args):
        self.changed = True
        return super().pop(*args)

    def __init__(self, value=None):
        if value is None:
            super().__init__()
        else:
            self.validate(value)
            super().__init__(value)
        self.changed = False

    # Methods related to selection and maniuplation by "criteria"
    # Numbers in "criteria" start with 1.

    def select(self, *criteria):
        """Take "criteria" (such as tags, ranges of numbers, etc) and return the
        indices of the resulting items. In other parts of the class and
        subclasses, operations need to know what items to act on.
        """
        selector = Selector(criteria)
        return [i for i in selector.indices(self.data)]

    def items(self, indices: list[int] = None):
        """Return the items for a set of indices"""
        if indices is None:
            return self.data
        else:
            return [self[i] for i in indices]

    def replace(self, indices: list, newvalues: list):
        """Replace existing items at the select provided.

        Also inserts if the indices run out. Does not create items; expects them
        to already exist. Does not return anything since the calling code
        already has the new items.
        """
        _indices = sorted(indices)
        _newvalues = newvalues.copy()
        max = (_indices[-1] + 1) if _indices else 0
        while _newvalues and _indices:
            self[_indices.pop(0)] = _newvalues.pop(0)
        while _indices:
            del self[_indices.pop()]
        self[max:max] = _newvalues

    def delete(self, indices: list[int]) -> list:
        """Remove the items at select and return them"""
        killlist, keeplist = self.split(indices)
        self.data = keeplist
        self.changed = True
        return killlist

    def split(self, indices: list[int]) -> tuple[list, list]:
        """Return a tuple of items at indices and items not so"""
        inlist = [t for i, t in enumerate(self) if i in indices]
        outlist = [t for i, t in enumerate(self) if i not in indices]
        return (inlist, outlist)

    def write_items(self, fileish: IOBase, indices: list = None):
        """
        Write all or a set of items to a file-like object
        """
        writer = DictWriter(fileish, self.schema, delimiter="|")
        for item in self.items(indices):
            values = dict([(f, getattr(item, f)) for f in self.schema])
            writer.writerow(values)

    def read_items(self, fileish: IOBase, indices: list = None):
        """
        Read all or a set of items from a file-like object
        """
        indices = indices if indices else []
        reader = DictReader(fileish, self.schema, delimiter="|")
        items = [Item(state=self.state, **d) for d in reader]
        self.replace(indices, items)
        return items
